/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/****************************
 * Loading via ImageMagick. *
 ****************************/

#include "gliv.h"
#include "math_floats.h"        /* sqrtf(), ceilf() */

#include <sys/stat.h>           /* S_IRWXU */
#include <sys/types.h>          /* pid_t */
#include <sys/wait.h>           /* waitpid() */
#include <stdio.h>              /* tmpnam(), remove(), perror(), FILE... */
#include <time.h>               /* time(), nanosleep() */
#include <unistd.h>             /* *dir(), getp*id(), fork(), _exit(), ... */
#include <stdlib.h>             /* system() */
#include <signal.h>             /* raise(), signal(), SIG* */
#include <string.h>             /* strstr() */
#include <regex.h>              /* ^reg.* */
#include <fcntl.h>              /* open(), O_*, S_*, fcntl(), F_GETFL  */

extern rt_struct *rt;

/*
 * Used to store the association between an unknown file to
 * gdk-pixbuf and its PNG equivalent made by ImageMagick.
 */
static GHashTable *hash = NULL;

/*
 * We chdir() before doing the ImageMagick stuff, because it can leave
 * some garbage if interrupted.  We will delete the dir at the end.
 */
static const gchar *gliv_wd, *magick_wd;

/* When the parent receives a SIGCHLD it is set to 0. */
static volatile pid_t child_pid = 0;

/* Used to know if we are in the parent or the child. */
static pid_t gliv_pid;

/* If we receive an exiting signal, we have to unitialize ImageMagick. */
static void signal_finish(gint sig)
{
    imagemagick_finish();
    signal(sig, SIG_DFL);
    raise(sig);
}

/* Called the first time we cannot load an image with gdk_pixbuf. */
static void imagemagick_init(void)
{
    gchar *tmp_dir;

    /* This hash table links images and their PNG equivalent. */
    hash = g_hash_table_new(g_str_hash, g_str_equal);

    gliv_wd = g_get_current_dir();

    magick_wd = g_strdup_printf("%s/gliv/%s_%ld_%d", g_get_tmp_dir(),
                                g_path_get_basename(tmpnam(NULL)),
                                time(NULL), getpid());

    tmp_dir = g_path_get_dirname(magick_wd);
    mkdir(tmp_dir, S_IRWXU);
    g_free(tmp_dir);

    mkdir(magick_wd, S_IRWXU);

    /* ImageMagick may want to put its temporary files there. */
    putenv(g_strconcat("TMPDIR=", magick_wd, NULL));
    putenv(g_strconcat("TEMP=", magick_wd, NULL));

    gliv_pid = getpid();

    signal(SIGHUP, signal_finish);
    signal(SIGINT, signal_finish);
    signal(SIGTERM, signal_finish);
    g_atexit(imagemagick_finish);
}

gboolean have_imagemagick(void)
{
    static gboolean checked = FALSE;
    static gboolean have_it;

    if (checked == FALSE) {
        /* First time. */

        have_it =
            !system("(convert  2>/dev/null | head -1 | grep -q ImageMagick) && \
                     (montage  2>/dev/null | head -1 | grep -q ImageMagick) && \
                     (identify 2>/dev/null | head -1 | grep -q ImageMagick)");

        checked = TRUE;
    }

    return have_it;
}

static void delete_png_tiles(const gchar * png_name, gint nb_tiles)
{
    gchar *filename;

    /* This is just for filename to have the correct length. */
    filename = g_strdup_printf("%s.%d", png_name, nb_tiles);

    while (nb_tiles > 0) {
        sprintf(filename, "%s.%d", png_name, --nb_tiles);

        if (remove(filename) < 0)
            perror(filename);
    }

    g_free(filename);
}

static void place_tiles(GdkPixbuf * pixbuf, const gchar * png_name, gint nb,
                        gint nb_x, gint tile_width, gint tile_height)
{
    gint i;
    GdkPixbuf *tile;
    gchar *filename;

    for (i = 0; i < nb; i++) {
        process_events();

        filename = g_strdup_printf("%s.%d", png_name, i);
        tile = load_gdk_pixbuf(filename);
        g_free(filename);

        process_events();

        gdk_pixbuf_copy_area(tile, 0, 0,
                             gdk_pixbuf_get_width(tile),
                             gdk_pixbuf_get_height(tile),
                             pixbuf,
                             (i % nb_x) * tile_width, (i / nb_x) * tile_height);

        gdk_pixbuf_unref(tile);
    }
}

static gboolean get_format(gint nb, gboolean * has_alpha,
                           gint * max_width, gint * max_height,
                           gint * nb_x, gint * nb_y)
{
    FILE *f;
    regex_t geom_regex, type_regex;
    gchar *buffer = NULL;
    gint size = 0, flags;
    regmatch_t match[3];

    *has_alpha = FALSE;
    *max_width = 0;
    *max_height = 0;

    f = fopen("id", "r");
    if (f == NULL)
        return FALSE;

    flags = REG_EXTENDED;
    regcomp(&geom_regex, "^ *Geometry: ([0-9]+)x([0-9]+)\n?$", flags);

    flags |= REG_NOSUB;
    regcomp(&type_regex, "^ *Type:.*with transparency\n?$", flags);

    while (!feof(f)) {
        getline(&buffer, &size, f);

        if (!*has_alpha && strstr(buffer, "with transparency") &&
            !regexec(&type_regex, buffer, 0, NULL, 0))

            *has_alpha = TRUE;

        else if (strstr(buffer, "Geometry: ") &&
                 !regexec(&geom_regex, buffer, 3, match, 0)) {
            gint w, h;

            buffer[match[1].rm_eo] = '\0';
            buffer[match[2].rm_eo] = '\0';

            w = atoi(buffer + match[1].rm_so);
            h = atoi(buffer + match[2].rm_so);

            if (w > *max_width)
                *max_width = w;

            if (h > *max_height)
                *max_height = h;
        }
    }

    regfree(&geom_regex);
    regfree(&type_regex);

    /* We want both functions to be called. */
    if ((fclose(f) < 0) | (remove("id") < 0))
        perror("id");

    /*
     * Number of tiles in each direction.
     * The logic is to have:
     *
     * nb_x * max_width    rt->scr_width
     * ----------------- ~ --------------
     * nb_y * max_height   rt->scr_height
     *
     * so that in maximize mode, the space usage and zoom ratio are maximal.
     */

    *nb_x = (gint) ceilf(sqrtf((gfloat) nb * *max_height * rt->scr_width /
                               (*max_width * rt->scr_height)));

    *nb_y = (gint) ceilf((gfloat) nb / *nb_x);

    return TRUE;
}

static gint count_tiles(const gchar * png_name)
{
    gint nb = 0;
    gchar *filename;
    gboolean found;

    do {
        filename = g_strdup_printf("%s.%d", png_name, nb);
        found = g_file_test(filename, G_FILE_TEST_EXISTS);
        g_free(filename);
        nb++;
    } while (found);

    return nb - 1;
}

/* We build the concatenation, save it and return it. */
static GdkPixbuf *png_concatenate(const gchar * png_name)
{
    gint nb, nb_x, nb_y, tile_width, tile_height;
    GdkPixbuf *pixbuf;
    GError *err = NULL;
    gboolean has_alpha;

    /* How many tiles? */
    nb = count_tiles(png_name);

    process_events();

    if (!get_format(nb, &has_alpha, &tile_width, &tile_height, &nb_x, &nb_y))
        return NULL;

    process_events();

    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, has_alpha, 8,
                            nb_x * tile_width, nb_y * tile_height);

    /* Blacken it. */
    gdk_pixbuf_fill(pixbuf, 0x00000000);

    process_events();

    place_tiles(pixbuf, png_name, nb, nb_x, tile_width, tile_height);

    process_events();

    if (gdk_pixbuf_save(pixbuf, png_name, "png", &err, NULL) == FALSE) {
        g_printerr("%s: %s\n", locale_to_utf8(png_name), err->message);
        g_error_free(err);
    }

    process_events();

    delete_png_tiles(png_name, nb);

    return pixbuf;
}

static gint wait_child(pid_t pid)
{
    const struct timespec req = {.tv_sec = 0,.tv_nsec = 10000000 };
    gint status;

    child_pid = pid;

    while (child_pid != 0) {
        /* Wait for SIGCHLD */
        process_events();
        nanosleep(&req, NULL);
    }

    /* The child process exited. */

    waitpid(pid, &status, WNOHANG);

    return WIFEXITED(status) ? WEXITSTATUS(status) : -1;
}

/* Return FALSE if failed. */
gboolean execute_command(gchar ** argv, gchar * out)
{
    pid_t pid;
    gint status;

    pid = fork();
    if (pid < 0)
        return FALSE;

    if (pid == 0) {
        if (out != NULL) {
            close(STDIN_FILENO);
            open("/dev/null", O_RDONLY);

            close(STDOUT_FILENO);
            open(out, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
        }

        return execvp(argv[0], argv);
    }

    if (waitpid(pid, &status, 0) < 0)
        return FALSE;

    return WIFEXITED(status) && WEXITSTATUS(status) == 0;
}

/* Return FALSE if failed. */
static gboolean make_id(const gchar * png_name, gint nb)
{
    gchar **argv;
    gint i, res;

    argv = g_new(gchar *, nb + 3);
    argv[0] = "identify";
    argv[1] = "-verbose";

    for (i = 0; i < nb; i++)
        argv[i + 2] = g_strdup_printf("%s.%d", png_name, i);

    argv[nb + 2] = NULL;

    res = execute_command(argv, "id");

    for (i = 0; i < nb; i++)
        g_free(argv[i + 2]);

    g_free(argv);
    return res;
}

G_GNUC_NORETURN static void child_process(const gchar * filename,
                                          gchar * png_name)
{
    gchar **argv;
    const gchar *dir;
    gboolean failed;

    dir = g_path_is_absolute(filename) ? "" : gliv_wd;

    argv = g_new(gchar *, 4);
    argv[0] = "convert";
    argv[1] = fast_build_filename(dir, filename);
    argv[2] = png_name;
    argv[3] = NULL;

    /* The conversion. */
    failed = !execute_command(argv, NULL);

    g_free(argv[1]);
    g_free(argv);

    if (failed == FALSE && g_file_test(png_name, G_FILE_TEST_EXISTS) == FALSE) {
        failed = !make_id(png_name, count_tiles(png_name));
    }

    /* We don't want the child process to run the atexit() function. */
    _exit(failed);
}

static void sig_chld(gint unused)
{
    child_pid = 0;
}

static GdkPixbuf *convert_to_png(const gchar * filename, gchar ** png_name)
{
    pid_t pid;

    signal(SIGCHLD, sig_chld);
    *png_name = g_strdup_printf("%u_%s.png", g_random_int_range(0, 1 << 30),
                                g_path_get_basename(filename));

    pid = fork();

    if (pid < 0) {
        perror("fork");
        return NULL;

    } else if (pid == 0)
        child_process(filename, *png_name);

    else {
        /* Parent. */
        if (wait_child(pid) != 0)
            return NULL;

        if (g_file_test(*png_name, G_FILE_TEST_EXISTS) == FALSE)
            /* Multipart image. */
            return png_concatenate(*png_name);
    }

    return load_gdk_pixbuf(*png_name);
}

GdkPixbuf *load_with_imagemagick(const gchar * filename)
{
    gchar *png_name;
    GdkPixbuf *pixbuf;

    if (have_imagemagick() == FALSE) {
        ERROR_MSG(_("%s: ImageMagick not found\n"), locale_to_utf8(filename));
        return NULL;
    }

    if (hash == NULL)
        /* First time. */
        imagemagick_init();

    chdir(magick_wd);

    png_name = g_hash_table_lookup(hash, filename);

    if (png_name == NULL) {
        /* First time this image is loaded. */
        pixbuf = convert_to_png(filename, &png_name);

        if (pixbuf != NULL)
            /* The conversion succeeded. */
            g_hash_table_insert(hash, (gchar *) filename, png_name);

    } else {
        /* This image has already been converted. */
        pixbuf = load_gdk_pixbuf(png_name);

        if (pixbuf == NULL)
            /*
             * It can happen if the png image could
             * not be saved when it was created.
             */
            pixbuf = convert_to_png(filename, &png_name);
    }

    chdir(gliv_wd);

    return pixbuf;
}

static void suppress_file(gpointer unused, gpointer filename)
{
    if (remove(filename) < 0)
        perror(filename);
}

/*
 * Delete the temporary directory we created.
 * This function is called on exit so there is
 * no need freeing memory.
 */
void imagemagick_finish(void)
{
    gchar *tmp_dir, *path, *message;
    FILE *f;

    if (getppid() == gliv_pid) {
        ERROR_MSG(_("Warning: the ImageMagick process is still running\n"));
        return;
    }

    if (hash == NULL)
        return;

    if (child_pid != 0)
        kill(child_pid, SIGTERM);

    /* Delete converted images. */
    chdir(magick_wd);
    g_hash_table_foreach(hash, (GHFunc) suppress_file, NULL);
    hash = NULL;
    chdir(gliv_wd);

    /* Delete the gliv temporary directory. */
    tmp_dir = g_path_get_dirname(magick_wd);
    if (!rmdir(magick_wd) && !rmdir(tmp_dir))
        /* OK. */
        return;

    path = fast_build_filename(tmp_dir, "README");
    if (g_file_test(path, G_FILE_TEST_EXISTS))
        return;

    f = fopen(path, "w");
    if (f == NULL) {
        perror(path);
        return;
    }

    message = _("This directory can be deleted if GLiv is not running");
    fprintf(f, "%s\n", message);
    fclose(f);
}
