/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/************************
 * Which image to load. *
 ************************/

#include "gliv.h"

extern rt_struct *rt;
extern options_struct *options;
extern gliv_image *current_image;

static gliv_image *previous_image = NULL;
static gliv_image *next_image = NULL;

static GtkMenuItem *slide_show_menu;
static guint slide_show_timer = 0;
static gboolean slide_show_immediately = FALSE;

/*
 * Once the images list has been reordered we must take care
 * of the current image position and destroy the previous and
 * next images since they changed.
 */
void after_reorder(gint new_id)
{
    if (previous_image != NULL &&
        current_image->node->prev != previous_image->node) {

        destroy_image(previous_image);
        previous_image = NULL;
    }

    if (next_image != NULL && current_image->node->next != next_image->node) {
        destroy_image(next_image);
        next_image = NULL;
    }

    current_image->number = new_id;
    fill_ident(current_image);

    refresh(REFRESH_STATUS);

    g_free(current_image->ident);
    current_image->ident = NULL;
}

/*** Next/previous image ***/

static GList *next_node(GList * node, gint dir)
{
    if (node == NULL)
        return (dir == 1) ? get_list_head() : get_list_end();

    if (dir == 1)
        return node->next;

    /* dir == -1 */
    return node->prev;
}

static gliv_image *load_next_node(GList ** node, gint dir)
{
    gliv_image *im = NULL;
    GList *next;

    while ((next = next_node(*node, dir)) != NULL) {
        im = load_file(next->data);
        if (im == NULL)
            /* Delete bad images from the list. */
            remove_from_list(next);
        else {
            *node = next;
            break;
        }
    }

    return im;
}

static gliv_image *load_in_direction(gint dir)
{
    gliv_image *im = NULL;
    GList *node;
    gint id;

    if (current_image == NULL) {
        /* First time. */
        node = NULL;
        id = 0;
    } else {
        node = current_image->node;
        id = current_image->number + dir;
    }

    if (get_list_length() == 2 && options->loop) {
        /*
         * When there are only two images, and we are looping,
         * next_image and previous_image are equal.
         */

        if (previous_image != NULL)
            return previous_image;

        if (next_image != NULL)
            return next_image;
    }


    im = load_next_node(&node, dir);

    if (im == NULL && options->loop && get_list_length() > 1) {
        node = NULL;
        id = (dir == 1) ? 0 : (get_list_length() - 1);
        im = load_next_node(&node, dir);
    }

    if (im != NULL) {
        im->node = node;
        im->number = id;
    }

    return im;
}

/* Return FALSE if there are no more images. */
gboolean load_direction(gint dir)
{
    gliv_image **backup;
    gliv_image **next;
    gboolean warp;
    gboolean no_next;

    static gint next_direction;

    /*
     * Before loading the next image in a prebuffer we let GTK process
     * remaining events and if there is a request for the next
     * image in the events pending then -> crash.
     * With this boolean we process loading request only if we know
     * all loadings are finished.
     * IOW: we are not reentrant.
     */
    static gboolean load_finished = TRUE;

    if (load_finished == FALSE) {
        /*
         * We come from a gtk_main_iteration_do(FALSE) called here,
         * so we stop processing events and we load the next image.
         */
        next_direction = dir;
        return TRUE;
    }

    if (is_loading() || current_image == NULL)
        return TRUE;

    load_finished = FALSE;

    while (dir * dir == 1) {

        no_next = (dir == 1 && current_image->node->next == NULL) ||
            (dir == -1 && current_image->node->prev == NULL);

        warp = no_next && options->loop && get_list_length() > 1;

        if (warp == FALSE && no_next) {
            /* Out of bounds. */
            load_finished = TRUE;
            return FALSE;
        }

        if (dir == -1) {
            backup = &next_image;
            next = &previous_image;
        } else {
            /* dir == 1 */
            backup = &previous_image;
            next = &next_image;
        }

        next_direction = 0;

        /*
         * If there are only two images,
         * we can have *backup == *next.
         */
        if (*backup != *next) {
            destroy_image(*backup);
            *backup = NULL;
        }

        prioritize_textures(current_image, FALSE);

        if (*next == NULL)
            *next = load_in_direction(dir);

        *backup = current_image;
        current_image = *next;
        *next = NULL;

        render();

        while (gtk_events_pending() != 0 && next_direction == 0)
            gtk_main_iteration_do(FALSE);

        *next = load_in_direction(dir);

        /* Usually next_direction == 0, and the while loop ends. */
        dir = next_direction;
    }

    load_finished = TRUE;
    return TRUE;
}

/*
 * After many images are opened we want to load
 * the next image but not the preloaded one.
 */
void open_next_image(void)
{
    if (is_loading())
        return;

    destroy_image(next_image);
    destroy_image(previous_image);

    prioritize_textures(current_image, FALSE);
    previous_image = current_image;

    current_image = load_in_direction(1);
    render();

    next_image = load_in_direction(1);
}

/*** First/second image ***/

void load_first_image(void)
{
    install_segv_handler();
    init_history();

    if (get_list_head() == NULL)
        return;

    current_image = load_in_direction(1);
    if (current_image == NULL) {
        ERROR_MSG(_("No image found\n"));
        quit(1);
    }

    render();
}

void load_second_image(void)
{
    if (slide_show_immediately)
        start_slide_show();

    else if (next_image == NULL)
        next_image = load_in_direction(1);
}

/*** Random loading ***/

void open_file(const gchar * filename, gboolean add_dir, gboolean shuffle_flag)
{
    gchar *dir;
    gliv_image *im;

    if (is_loading())
        return;

    /* Try to load the requested image. */
    im = load_file(filename);
    if (im == NULL)
        return;

    if (add_dir) {
        /* Add all files in the directory in the list. */
        dir = g_path_get_dirname(filename);

        insert_after_current(&dir, 1, shuffle_flag);
        g_free(dir);

        /* Put the requested image at current+1 position in the file list. */
        place_next(filename);
    } else {
        gchar *tmp;

        tmp = g_strdup(filename);
        insert_after_current(&tmp, 1, shuffle_flag);
        g_free(tmp);
    }

    if (current_image == NULL) {
        im->node = get_list_head();
        im->number = 0;
    } else {
        im->node = current_image->node->next;
        im->number = current_image->number + 1;
    }

    destroy_image(previous_image);
    destroy_image(next_image);

    prioritize_textures(current_image, FALSE);
    previous_image = current_image;
    current_image = im;

    render();

    next_image = load_in_direction(1);
}

/*
 * Reloading.
 * Used for example, when an option is changed such as dithering or mipmap.
 */

static gliv_image *reload_image(gliv_image * im)
{
    gliv_image *new;

    if (im == NULL)
        return NULL;

    new = load_file(im->node->data);
    if (new == NULL)
        return im;

    new->node = im->node;
    new->number = im->number;

    destroy_image(im);

    return new;
}

void reload_all_images(void)
{
    if (is_loading())
        return;

    current_image = reload_image(current_image);
    prioritize_textures(current_image, TRUE);

    rt->filtering_enabled = FALSE;
    refresh(REFRESH_NOW);

    previous_image = reload_image(previous_image);
    next_image = reload_image(next_image);
}

/*** Menu ***/

static gboolean check_direction(const gchar * filename, gint dir)
{
    gliv_image **prev, **next;
    GList *next_node;

    if (dir == 1) {
        prev = &previous_image;
        next = &next_image;

        next_node = current_image->node->next;
        if (next_node != NULL)
            next_node = next_node->next;
    } else {
        /* dir == -1 */
        prev = &next_image;
        next = &previous_image;

        next_node = current_image->node->prev;
        if (next_node != NULL)
            next_node = next_node->prev;
    }

    if (*next != NULL) {

        /* Check if the requested image is just following the current one. */
        if (g_str_equal(filename, (*next)->node->data)) {
            load_direction(dir);
            return TRUE;
        }

        if (next_node != NULL) {
            /*
             * Check if the requested image is just before the previous one,
             * or just after the next one.
             */
            if (g_str_equal(filename, next_node->data)) {
                gint number;

                destroy_image(*prev);
                *prev = *next;
                *next = NULL;

                number = current_image->number + dir;
                destroy_image(current_image);
                current_image = load_file(filename);
                current_image->node = next_node;
                current_image->number = number;

                render();
                return TRUE;
            }
        }
    }

    return FALSE;
}

void menu_load(const gchar * filename)
{
    GList *node;
    gint number = 0;
    gliv_image *im;
    gchar *clean;

    clean = clean_filename(filename);

    if (is_loading() ||
        g_str_equal(clean, current_image->node->data) ||
        check_direction(clean, -1) || check_direction(clean, 1)) {

        g_free(clean);
        return;
    }

    for (node = get_list_head(); node != NULL; node = node->next) {
        if (g_str_equal(clean, node->data))
            break;

        number++;
    }

    if (node == NULL) {
        /* The image is not in the list anymore. */
        g_free(clean);
        return;
    }

    im = load_file(clean);
    g_free(clean);
    if (im == NULL) {
        remove_from_list(node);
        return;
    }

    destroy_image(previous_image);
    previous_image = NULL;

    destroy_image(next_image);
    next_image = NULL;

    current_image = im;
    current_image->node = node;
    current_image->number = number;

    render();
}

#define TRY_UNLOAD(image)                       \
    if (image != NULL && image->node == node) { \
        destroy_image(image);                   \
        image = NULL;                           \
    }

void unload(GList * node)
{
    TRY_UNLOAD(previous_image);
    TRY_UNLOAD(next_image);
}

/*** Slide show ***/

void set_slide_show_menu(GtkMenuItem * item)
{
    slide_show_menu = item;
}

static void stop_slide_show(void)
{
    if (slide_show_started()) {
        g_source_remove(slide_show_timer);
        slide_show_timer = 0;
    }

    set_menu_label(slide_show_menu, _("Start the slide show"), TRUE);
}

static gboolean slide_show_next(void)
{
    if (load_direction(1) == FALSE) {
        stop_slide_show();
        return FALSE;
    }

    return TRUE;
}

void start_slide_show(void)
{
    stop_slide_show();

    if (options->duration < 0 || current_image == NULL)
        return;

    slide_show_timer = g_timeout_add(options->duration * 1000,
                                     (GSourceFunc) slide_show_next, NULL);

    set_menu_label(slide_show_menu, _("Stop the slide show"), TRUE);
}

gboolean slide_show_started(void)
{
    return slide_show_timer != 0;
}

void toggle_slide_show(void)
{
    if (slide_show_started())
        stop_slide_show();
    else
        start_slide_show();
}

void begin_with_slide_show(void)
{
    slide_show_immediately = TRUE;
}
