/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/***********************
 * Configuration file. *
 ***********************/

#include "gliv.h"

#ifndef HAVE_GETLINE
#   include "../lib/getline.h"
#endif

#include <unistd.h>             /* R_OK */
#include <stdio.h>              /* FILE, f*(), perror(), getline() */
#include <string.h>             /* strlen(), memcpy() */

static GHashTable *table = NULL;
static gboolean options_read = FALSE;

/* The link between options and the rcfile. */
static options_struct opts = {
/* Default options */
    .fullscreen = FALSE,
    .maximize = FALSE,
    .scaledown = FALSE,
    .menu_bar = TRUE,
    .status_bar = TRUE,
    .scrollbars = TRUE,
    .zoom_pointer = FALSE,
    .alpha_checks = TRUE,
    .dither = FALSE,
    .force = FALSE,
    .build_menus = FALSE,
    .mipmap = FALSE,
    .mnemonics = FALSE,
    .loop = FALSE,
    .delay = 0,
    .history_size = 1000,
    .duration = 10,
    .bg_col = {0, 0, 0},
    .alpha1 = {0x6666, 0x6666, 0x6666},
    .alpha2 = {0x9999, 0x9999, 0x9999}
};

typedef struct {
    const gchar *name;
    gint *option;
    const gchar *comment;
    const gboolean is_bool;
} option_struct;

/* *INDENT-OFF* */
static option_struct option_names[] = {
/* To fill the hash table and the configuration file. */
{ "full-screen",  &opts.fullscreen,   N_("Start in full screen mode"),      1 },
{ "maximize",     &opts.maximize,     N_("Maximize small images"),          1 },
{ "scale-down",   &opts.scaledown,    N_("Scale down larges images"),       1 },
{ "menu",         &opts.menu_bar,     N_("Display the menu bar"),           1 },
{ "info",         &opts.status_bar,   N_("Display infos about the image"),  1 },
{ "scrollbars",   &opts.scrollbars,   N_("Display scrollbars"),             1 },
{ "zoom-pointer", &opts.zoom_pointer, N_("Zoom centered on pointer"),       1 },
{ "alpha-checks", &opts.alpha_checks, N_("Alpha checks in the background"), 1 },
{ "dither",       &opts.dither,       N_("Dithering"),                      1 },
{ "force-load",   &opts.force,        N_("Try to load every file"),         1 },
{ "build-menus",  &opts.build_menus,  N_("Build images menus at startup"),  1 },
{ "mipmap",       &opts.mipmap,       N_("Build Mipmaps"),                  1 },
{ "mnemonics",    &opts.mnemonics,    N_("Make mnemonics for images menus"),1 },
{ "loop",         &opts.loop,         N_("Make the slide show loop"),       1 },
{ "delay",        &opts.delay,        N_("Delay before hiding the cursor"), 0 },
{ "history",      &opts.history_size, N_("History length"),                 0 },
{ "slide-show",   &opts.duration,     N_("Delay between images"),           0 },
{ "bg_col_red",   opts.bg_col,        N_("background: red channel"),        0 },
{ "bg_col_green", opts.bg_col + 1,    N_("background: green channel"),      0 },
{ "bg_col_blue",  opts.bg_col + 2,    N_("background: blue channel"),       0 },
{ "alpha1_red",   opts.alpha1,        N_("alpha1 tile: red channel"),       0 },
{ "alpha1_green", opts.alpha1 + 1,    N_("alpha1 tile: green channel"),     0 },
{ "alpha1_blue",  opts.alpha1 + 2,    N_("alpha1 tile: blue channel"),      0 },
{ "alpha2_red",   opts.alpha2,        N_("alpha2 tile: red channel"),       0 },
{ "alpha2_green", opts.alpha2 + 1,    N_("alpha2 tile: green channel"),     0 },
{ "alpha2_blue",  opts.alpha2 + 2,    N_("alpha2 tile: blue channel"),      0 },
{ NULL,           NULL,               NULL,                                 0 }
};
/* *INDENT-ON* */

/*
 * Maximum length of the option names, currently
 * it is strlen("zoom-pointer") == 12.
 * Used to indent the option file.
 */
#define MAX_OPT_LEN 12

/* Between the options and the keyboard accelerators. */
#define SEPARATOR "==========\n"

static gchar *find_rcfile(gboolean test)
{
    gchar *filename;

    filename = g_build_filename(g_get_home_dir(), ".glivrc", NULL);

    if (test == FALSE || g_file_test(filename, G_FILE_TEST_EXISTS))
        /* When loading the file, we really try to access it. */
        /* When saving, we just need the filename. */
        return filename;

    g_free(filename);
    return NULL;
}

/*** Loading options. ***/

static void init_hash_table(void)
{
    gint i;

    table = g_hash_table_new(g_str_hash, g_str_equal);

    for (i = 0; option_names[i].name != NULL; i++)
        g_hash_table_insert(table, (gchar *) option_names[i].name,
                            &option_names[i]);
}

 /* Processes spaces and '#'. */
static gchar *clean_str(const gchar * str)
{
    gchar *new_str;
    gchar *ptr;

    new_str = g_new(gchar, strlen(str) + 1);

    for (ptr = new_str; *str != '\n' && *str != '\0'; str++) {
        if (*str == ' ')
            continue;

        if (*str == '#') {
            g_free(new_str);
            return NULL;
        }

        *ptr = *str;
        ptr++;
    }
    *ptr = '\0';

    new_str = g_renew(gchar, new_str, ptr - new_str + 1);
    return new_str;
}

static void process_line(const gchar * line)
{
    gchar **res;
    option_struct *opt;
    gchar *clean;

    if (*line == '\n' || *line == '\0')
        /* Skip this line. */
        return;

    clean = clean_str(line);
    if (clean == NULL)
        /* This line is a '#' comment. */
        return;

    /* res[0]: option name ; res[1]: value */
    res = g_strsplit(clean, "=", 2);

    g_free(clean);

    if (res[0] != NULL && res[1] != NULL && res[2] == NULL) {
        /* No error during split. */
        opt = g_hash_table_lookup(table, res[0]);

        if (opt != NULL) {
            /* Option found. */

            if (opt->is_bool) {
                gboolean *bool;

                bool = (gboolean *) opt->option;
                *bool = g_strcasecmp(res[1], "true") ? FALSE : TRUE;
            } else
                /* opt->is_bool == FALSE */
                *opt->option = (gint) g_strtod(res[1], NULL);
        }
    }

    g_strfreev(res);
}

options_struct *load_rc(gboolean from_file)
{
    gchar *filename;
    FILE *file;
    gchar *line = NULL;
    size_t nb = 0;

    if (from_file == FALSE)
        return &opts;

    init_hash_table();

    filename = find_rcfile(TRUE);
    if (filename == NULL)
        return &opts;

    file = fopen(filename, "r");
    if (file == NULL) {
        /*
         * Not very frequent because find_rcfile(TRUE)
         * should have made us return.
         */
        perror(filename);
        return &opts;
    }

    while (feof(file) == 0) {
        getline(&line, &nb, file);
        if (g_str_equal(line, SEPARATOR))
            break;

        process_line(line);
    }

    g_free(line);
    fclose(file);
    g_free(filename);
    g_hash_table_destroy(table);

    options_read = TRUE;
    return &opts;
}

void load_accelerators(void)
{
    gchar *filename;
    FILE *file;
    gchar *line = NULL;
    size_t nb = 0;

    if (options_read == FALSE)
        return;

    filename = find_rcfile(TRUE);
    if (filename == NULL)
        return;

    file = fopen(filename, "r");
    if (file == NULL) {
        perror(filename);
        return;
    }

    while (getline(&line, &nb, file) != -1 &&
           g_str_equal(line, SEPARATOR) == FALSE);

    fflush(file);

    gtk_accel_map_load_fd(fileno(file));

    fclose(file);
}


/*** Saving options. ***/

static void write_option_line(FILE * f, gint index)
{
    gint i, value;

    fputs(option_names[index].name, f);

    for (i = strlen(option_names[index].name); i < MAX_OPT_LEN; i++)
        fputc(' ', f);

    value = *(option_names[index].option);

    if (option_names[index].is_bool)
        fprintf(f, " = %s\n\n", value ? "True" : "False");
    else
        fprintf(f, " = %d\n\n", value);
}

static void write_rc(gchar * filename)
{
    FILE *file;
    gint i;

    file = fopen(filename, "w");
    if (file == NULL) {
        perror(filename);
        return;
    }

    fprintf(file, _("# Configuration file for GLiv %s\n\n"), VERSION);
    fputs(_("# Option names are case sensitive.\n"), file);
    fputs(_("# Option values are case insensitive.\n\n"), file);
    fputs(_("# Note: 'maximize = True' or 'scale-down = True'"
            " implies 'full-screen = True'.\n\n"), file);

    for (i = 0; option_names[i].name != NULL; i++) {
        /* The comment line. */
        fprintf(file, "# %s\n", _(option_names[i].comment));

        /* The option line. */
        write_option_line(file, i);
    }

    /* The separation between options and keyboard accelerators. */
    fprintf(file, "%s\n", SEPARATOR);
    fflush(file);

    gtk_accel_map_save_fd(fileno(file));

    fclose(file);
    g_free(filename);
}

void save_rc(options_struct * opt)
{
    memcpy(&opts, opt, sizeof(options_struct));
    write_rc(find_rcfile(FALSE));
}
