/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/************************
 * Rendering functions. *
 ************************/

#include "gliv.h"

#include <gtk/gtkgl.h>
#include <GL/gl.h>

extern rt_struct *rt;
extern options_struct *options;
extern gliv_image *current_image;
extern GtkWidget *gl_widget;

static guint idle_id = 0;

static void set_filter(gint filter)
{
    gint level, id;
    texture_map *map;

    /* All textures have the same parameters. */

    for (level = 0; level < current_image->nb_maps; level++) {
        map = current_image->maps + level;

        for (id = 0; id < map->nb_tiles; id++) {
            glBindTexture(GL_TEXTURE_2D, map->tex_ids[id]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
        }
    }
}

static void call_lists(gboolean smooth, gint level)
{
    gint id;
    texture_map *map;

    if (rt->filtering_enabled != smooth) {
        rt->filtering_enabled = smooth;
        set_filter(smooth ? GL_LINEAR : GL_NEAREST);
    }

    map = current_image->maps + level;

    for (id = 0; id < map->nb_tiles; id++)
        if (matrix_tile_visible(map->tiles + id))
            glCallList(map->list + id);
}

static void draw_checker(void)
{
    static guint tex_id = 0;
    gint i;
    gfloat half_w, half_h;
    gboolean matrix_changed;

    matrix_changed = get_matrix_has_changed();

    if (rt->alpha_checks_changed) {
        gushort texture[12];

        if (tex_id == 0)
            glGenTextures(1, &tex_id);

        for (i = 0; i < 3; i++) {
            texture[i] = texture[9 + i] = options->alpha1[i];
            texture[3 + i] = texture[6 + i] = options->alpha2[i];
        }

        glBindTexture(GL_TEXTURE_2D, tex_id);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 2, 2, 0, GL_RGB,
                     GL_UNSIGNED_SHORT, texture);

        rt->alpha_checks_changed = FALSE;
    } else
        glBindTexture(GL_TEXTURE_2D, tex_id);

    if (matrix_changed == FALSE)
        /* glMatrixMode(GL_MODELVIEW); */
        /* Save the matrix only if we will not replace it when redrawing. */
        glPushMatrix();

    glLoadIdentity();

    half_w = rt->wid_size->width / 2.0;
    half_h = rt->wid_size->height / 2.0;

    glBegin(GL_QUADS);

    glTexCoord2f(0.0, 0.0);
    glVertex2f(-half_w, -half_h);

    glTexCoord2f(half_w / 16.0, 0.0);
    glVertex2f(half_w, -half_h);

    glTexCoord2f(half_w / 16.0, half_h / 16.0);
    glVertex2f(half_w, half_h);

    glTexCoord2f(0.0, half_h / 16.0);
    glVertex2f(-half_w, half_h);

    glEnd();

    if (matrix_changed == FALSE)
        /* glMatrixMode(GL_MODELVIEW); */
        glPopMatrix();
}

static gint choose_mipmap_level(void)
{
    gfloat zoom, mipmap_ratio = MIPMAP_RATIO;
    gint level = 1;

    if (options->mipmap) {
        zoom = get_matrix_zoom();

        while (mipmap_ratio > zoom && level < current_image->nb_maps) {
            mipmap_ratio *= MIPMAP_RATIO;
            level++;
        }
    }

    /* Mipmaps should only be scaled down. */
    return level - 1;
}

static gboolean redraw(void)
{
    GdkGLDrawable *gldrawable;

    clear_zoom_frame();

    gldrawable = gtk_widget_get_gl_drawable(gl_widget);
    gdk_gl_drawable_wait_gdk(gldrawable);

    glDisable(GL_DITHER);

    if (current_image != NULL && current_image->has_alpha &&
        options->alpha_checks)

        draw_checker();
    else
        glClear(GL_COLOR_BUFFER_BIT);

    glEnable(GL_DITHER);

    write_gl_matrix();

    if (current_image != NULL)
        call_lists(is_filtering_needed(), choose_mipmap_level());

    gdk_gl_drawable_swap_buffers(gldrawable);

    if (current_image != NULL && options->scrollbars)
        update_scrollbars(TRUE);

    gdk_gl_drawable_wait_gl(gldrawable);

    /* To remove the idle func. */
    idle_id = 0;
    return FALSE;
}

void render(void)
{
    prioritize_textures(current_image, TRUE);

    matrix_reset();
    delete_selected_image();
    fill_ident(current_image);

    rt->filtering_enabled = FALSE;

    if (options->fullscreen == FALSE)
        goto_window(FALSE);

    if (options->maximize || options->scaledown)
        matrix_set_max_zoom(-1, -1, TRUE);

    refresh(REFRESH_NOW | REFRESH_STATUS | APPEND_HISTORY);

    /* A bit dumb, but needed on some cards. */
    refresh(REFRESH_NOW);

    g_free(current_image->ident);
    current_image->ident = NULL;

    update_window_title();
    process_events();
}

void zoom_in(gfloat ratio)
{
    gint pointer_x, pointer_y;
    gfloat x, y;

    if (options->zoom_pointer) {
        /* The pointer is the zoom center. */
        gdk_window_get_pointer(gl_widget->window, &pointer_x, &pointer_y, NULL);
        x = (gfloat) pointer_x;
        y = (gfloat) pointer_y;
    } else {
        /* The zoom center is the midle of the window. */
        x = rt->wid_size->width / 2.0;
        y = rt->wid_size->height / 2.0;
    }

    matrix_zoom(ratio, x, y);
    refresh(REFRESH_IMAGE | REFRESH_STATUS | APPEND_HISTORY);
}

void refresh(gint what)
{
    /* GTK does that automatically but with more overhead. */
    if (what & REFRESH_IMAGE) {

        if (idle_id == 0)
            idle_id = gtk_idle_add_priority(GDK_PRIORITY_REDRAW,
                                            (GtkFunction) redraw, NULL);

    } else if (what & REFRESH_NOW) {

        if (idle_id != 0)
            gtk_idle_remove(idle_id);

        redraw();
    }

    if (what & REFRESH_STATUS)
        update_status_bar();

    if (what & APPEND_HISTORY)
        append_history();
}
