/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/*********************************
 * UTF-8 and filenames handling. *
 *********************************/

#include "gliv.h"

#include <string.h>             /* memcpy(), strlen(), strstr() */

/* The returned string should not be freed. */
const gchar *add_mnemonic(const gchar * str)
{
    static gchar *result = NULL;
    static gint size = 0;
    const gchar *ptr = NULL, *end;
    gint new_size;

    for (ptr = str; *ptr != '\0'; ptr = g_utf8_next_char(ptr))
        if (g_unichar_isalnum(g_utf8_get_char(ptr)))
            break;

    end = ptr + strlen(ptr);

    if (*ptr == '\0')
        /*
         * No g_unichar_isalnum() were found,
         * we add the underscore in front.
         */
        ptr = str;

    /* + 2: '_' and '\0'. */
    new_size = end - str + 2;
    if (new_size > size) {
        size = new_size;
        g_free(result);

        result = g_new(gchar, size);
    }

    memcpy(result, str, ptr - str);
    result[ptr - str] = '_';
    memcpy(result + (ptr - str) + 1, ptr, end - ptr + 1);

    return result;
}

static gboolean str_is_ascii(const gchar * str)
{
    while (*str != '\0') {
        if (*str >> 7)
            return FALSE;

        str++;
    }

    return TRUE;
}

/* The returned string should not be freed. */
const gchar *locale_to_utf8(const gchar * str)
{
    static gchar *result = NULL;
    GError *err = NULL;

    if (str_is_ascii(str))
        return str;

    g_free(result);

    result = g_locale_to_utf8(str, -1, NULL, NULL, &err);
    if (err != NULL) {
        g_printerr("%s\n", err->message);
        g_error_free(err);

        g_free(result);
        result = NULL;
        return str;
    }

    return result;
}

static gboolean is_clean(const gchar * filename)
{
    if (filename[0] != '/' && (filename[0] != '.' || filename[1] != '/'))
        return FALSE;

    return strstr(filename, "//") == NULL && strstr(filename, "/./") == NULL;
}

gchar *clean_filename(const gchar * filename)
{
    gchar *orig_copy, *copy, *ptr, *new, *new_ptr;
    gint count = 0, len = 0;
    gboolean absolute;

    if (is_clean(filename))
        return g_strdup(filename);

    /* We work on a copy as we may modify it. */
    orig_copy = copy = g_strdup(filename);

    if (copy[0] == '/') {
        /* Absolute filename. */
        while (copy[1] == '/')
            copy++;

        absolute = TRUE;
    } else {
        /* Relative filename. */
        gboolean finished = FALSE;

        while (finished == FALSE) {
            switch (copy[0]) {
            case '.':
                if (copy[1] == '/')
                    copy += 2;
                else
                    finished = TRUE;
                break;

            case '/':
                copy++;
                break;

            default:
                finished = TRUE;
            }
        }

        absolute = FALSE;
    }

    /*
     * "path1/./path2" is replaced with "path1///path2".
     * "./" and "//" are counted.
     */
    for (ptr = copy; *ptr != '\0'; ptr++) {
        len++;
        if (ptr[0] == '/')
            switch (ptr[1]) {
            case '.':
                if (ptr[2] == '/') {
                    ptr[1] = '/';
                    count++;
                }
                break;

            case '/':
                count++;
            }
    }

    if (count == 0) {
        /* The filename was clean. */

        if (absolute) {
            if (orig_copy != copy)
                /* The filename started with "//". */
                g_memmove(orig_copy, copy, len + 1);

            return orig_copy;
        }

        /* The relative filename just lacked the "./" in the beginning. */
        new = fast_build_filename(".", copy);
        g_free(orig_copy);
        return new;
    }

    /* We now have to remove the "//". */

    if (absolute) {
        new = g_new(gchar, len - count + 1);
        new_ptr = new;
    } else {
        new = g_new(gchar, len - count + 3);
        new[0] = '.';
        new[1] = '/';
        new_ptr = new + 2;
    }

    for (ptr = copy; *ptr != '\0'; ptr++)
        if (ptr[0] != '/' || ptr[1] != '/') {
            *new_ptr = *ptr;
            new_ptr++;
        }

    *new_ptr = '\0';

    g_free(orig_copy);

    return new;
}

gchar *clean_filename_free(gchar * filename)
{
    gchar *clean;

    if (filename[0] == '\0' || is_clean(filename))
        clean = filename;
    else {
        clean = clean_filename(filename);
        g_free(filename);
    }

    return clean;
}

gchar *fast_build_filename(const gchar * dir, const gchar * file)
{
    if (dir[0] == '\0')
        return g_strdup(file);

    if (file[0] == '\0')
        return g_strdup(dir);

    return g_strconcat(dir, "/", file, NULL);
}
